import sys
import numpy as np
from numpy import sin, cos, tan, pi, exp, sqrt

sys.path.append("c:/Programs/python/lib")
sys.path.append("d:/Programs/python/lib")

from tkargs import tkArgs
from tksci.tkoptimize import tkOptimize
from tkgraphic.tkplot import tkPlot


# target function
#functype = 'ellipsoid'
functype = 'general'


#algorism: simplex, sd, cg, newton, broyden, dfp, bfgs
algorism = 'simplex'
#algorism = 'sd'
#algorism = 'cg'
#algorism = 'newton'
#algorism = 'broyden'
#algorism = 'dfp'
#algorism = 'bfgs'

# for line search '' or 'none'/'newton', 'one', 'simple', 'exact', 'golden', 'armijo'
lsmode = 'armijo'


# optimization parameters
x0    = [0.0, 0.0]
optid = [  1,   1]
#for simplex
dx = np.array([3.0, 3.0])  # range to make an initial simplex


if len(sys.argv) >= 2:
    x0[0] = float(sys.argv[1])

if len(sys.argv) >= 3:
    x0[1] = float(sys.argv[2])

if len(sys.argv) >= 4:
    algorism = sys.argv[3]

if algorism == 'newton':
    lsmode = 'newton'

if len(sys.argv) >= 5:
    lsmode = sys.argv[4]

if len(sys.argv) >= 6:
    functype = sys.argv[5]


# for global optimization
nmaxiter = 200
tolx   = 1.0e-5
tolf   = 1.0e-5
iprintinterval = 1
print_level    = 4

# for Newton-Raphson method
dump  = 0.3

# line search parameters
ls_nmaxiter = 100

# for lsmode == 'exact'
ls_h = 1.0e-3  # x step to calculate first and second derivatives

# for lsmode == 'golden'
ls_alphaeps = 1.0e-2
ls_xrange = 0.5

if functype == 'ellipsoid':
    ls_alpha = 0.03
else:
    ls_alpha = 0.01

if lsmode == 'golden':
    ls_alpha = 1.0
elif lsmode == 'armijo':
    ls_alpha = 1.0


# graph plot parameters
fplot = 1
ngdata = 51
xgmin = -4.0
xgmax =  4.0
ygmin = -4.0
ygmax =  4.0
tsleep = 0.2

# file configuration
parameter_path = 'params.prm'
finalparameter_path = 'final_params.prm'


# first derivative of func(x)
def ediff1(i, x, optdata = None):
    if i == 0:
        return 2.0 * x[0]
    if i == 1:
        return 18.0 * x[1]

def gdiff1(i, x, optdata = None):
    if i == 0:
        return - 10.0 - 60.0 * x[0] + 4.5 * x[0]*x[0] + 12.0 * x[0]*x[0]*x[0] \
                + 3.0 * x[1] * x[1]
    if i == 1:
        return 30.0 - 60.0 * x[1] + 12.0 * x[1]*x[1]*x[1] \
                + 6.0 * x[0] * x[1]

# for Newton-Raphson method, second derivative of func(x)
def ediff2(i, j, x, optdata = None):
    if i == 0 and j == 0:
        return 2.0
    if i == 0 and j == 1:
        return 0.0
    if i == 1 and j == 0:
        return 0.0
    if i == 1 and j == 1:
        return 18.0

def gdiff2(i, j, x, optdata = None):
    if i == 0 and j == 0:
        return -60.0 + 9.0 * x[0] + 36.0 * x[0]*x[0]
    if i == 0 and j == 1:
        return 6.0 * x[1]
    if i == 1 and j == 0:
        return 6.0 * x[1]
    if i == 1 and j == 1:
        return -60.0 +36.0 * x[1]*x[1] + 6.0 * x[0]

def efunc(x, optdata = None):
    return x[0]*x[0] + 9.0*x[1]*x[1]

def gfunc(x, optdata = None):
    return -3.0 - 10.0 * x[0] - 30.0 * x[0]*x[0] + 1.5 * x[0]*x[0]*x[0] + 3.0 * x[0]*x[0]*x[0]*x[0] \
                + 30.0 * x[1] - 30.0 * x[1]*x[1] + 3.0 * x[1]*x[1]*x[1]*x[1] \
                + 3.0 * x[0] * x[1] * x[1]

# global function variables
if functype == 'ellipsoid':
    diff1 = ediff1
    diff2 = ediff2
    func  = efunc
else:
    diff1 = gdiff1
    diff2 = gdiff2
    func  = gfunc

# callback function. return code > 0 indicates error and simplex iteration will be terminated
def cfunc(optdata):
#    print("callback at iter={}".format(optdata.iter))
    x0 = optdata.RecoverParams(optdata.x0, optdata.x0_all, optdata.optid)
    if optdata.fplot == 1:
        graph = optdata.graph
        ax    = graph.axes[0]
        ax2   = graph.axes[1]
        if optdata.method == 'simplex':
            vtx = optdata.vtx
            x = []
            y = []
            for i in range(optdata.nvtx):
                x.append(vtx[i].x[0])
                y.append(vtx[i].x[1])
            x.append(vtx[0].x[0])
            y.append(vtx[0].x[1])
            graph.plot(0, x, y, color = 'blue', marker = '', linestyle = '-')
        else:
            xt = ax.xlist[0]
            yt = ax.ylist[0]
            xt.append(x0[0])
            yt.append(x0[1])
            graph.set_data(0, 0, 0, xt, yt)

        graph.pause(optdata.tsleep)
#        graph.sleep(optdata.tsleep)
    return 0


def main():
    print("Find minimum point by steepest-descend / conjugate gradient methods")
    print("")

    opt = tkOptimize()
    opt.add_parameter("x", x0[0], dx[0], optid[0])
    opt.add_parameter("y", x0[1], dx[1], optid[1])
    opt.set_method(algorism, lsmode)
    opt.set_functions(fitfunc = func, func = func, diff1func = diff1, diff2func = diff2)
    opt.initialize(callback = cfunc)
#    opt.initialize(
#            nmaxiter = nmaxiter, tolx = tolx, tolf = tolf,
#            callback = cfunc, print_level = print_level, iprintinterval = iprintinterval,
#            dump = dump,
#            ls_xrange = ls_xrange, ls_h = ls_h, ls_alpha = ls_alpha, ls_dump = dump, 
#            ls_nmaxiter = ls_nmaxiter, ls_alphaeps = ls_alphaeps
#            )
    var = opt.make_optdata(tsleep = tsleep)

#    inf = opt.read_parameters(parameter_path, update_params = 0)
#    print("parameters in [{}]".format(parameter_path))
#    for key, val in inf.items():
#        print("{}: {}".format(key, val))
#    print("save to [{}]".format(parameter_path))
#    opt.save_parameters(parameter_path)
#    exit()


    print("")
    print("Initial parameters:")
    opt.print_parameters()


# plot surface graph
    if fplot == 1:
        xgstep = (xgmax - xgmin) / (ngdata - 1)
        ygstep = (ygmax - ygmin) / (ngdata - 1)
        xg = np.empty([ngdata, ngdata])
        yg = np.empty([ngdata, ngdata])
        zg = np.empty([ngdata, ngdata])
        for ix in range(ngdata):
            for iy in range(ngdata):
                xg[ix, iy] = xgmin + ix * xgstep
                yg[ix, iy] = ygmin + iy * ygstep
                zg[ix, iy] = func([xg[ix][iy], yg[ix][iy]])

        graph = tkPlot(figsize = (10, 5))
        graph.add_subplot(1, 2, 0)
        graph.add_subplot(1, 2, 1, projection = '3d')

        f = func(x0)
        print("x0 = ({}, {}): f = {}".format(x0[0], x0[1], f))
        xt = []
        yt = []
        xt.append(x0[0])
        yt.append(x0[1])
        graph.plot(0, xt, yt, color = 'blue', linestyle = '-', linewidth = 0.5,
                            fillstyle = 'full', marker = 'o', markersize = 5)
        graph.plot_contour(0, xg, yg, zg, levels = 51, cmap = 'Spectral')
        graph.set_axtitle(0, 'contour')

        graph.plot_wireframe(1, xg, yg, zg, rstride = 2, cstride = 2)

        graph.pause()
#        graph.show()

        var.graph   = graph
        var.fplot   = fplot


# calculate initial parameters
    xmin, fmin, optdata = opt.optimize()
    print("")
    print("Optimized parameters:")
    opt.print_parameters(x0 = xmin, f = fmin)

    if opt.save_parameters(finalparameter_path):
        print("")
        print("Final resuls are stored to [{}]".format(finalparameter_path))
    else:
        print("")
        print("Error: Can not save the results to [{}]".format(finalparameter_path))


    if fplot == 1:
        print("Press enter to terminate:", end = '')
        ret = input()


if __name__ == "__main__":
    main()

